###############################################################################
# Copyright (c) 2006-2016 Franz Inc.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
###############################################################################

"""
A pycurl-based implementation of HTTP backend.
"""

from __future__ import print_function

import pycurl

import os
import threading
import time
import io

import errno

from past.builtins import unicode, basestring
from future.utils import iteritems

from franz.openrdf.util.strings import to_native_string, to_bytes

# Public symbols
__all__ = ['makeRequest']

# A simple way of checking if we're on pycurl>=7.19.3
if hasattr(pycurl, 'NOPROXY'):
    copt = lambda x: x
else:
    def copt(value):
        """
        Convert a value to a type suitable for a curl option.
        This simply means that unicode strings must be encoded
        to byte strings in older versions of curl.
        """ 
        if isinstance(value, unicode):
            return value.encode('ascii')
        return value          
    

curlPool = None

class Pool(object):
    """
    A pool of curl objects.

    Curl instances are not thread-safe, so we maintain this
    pool to avoid the need of creating a fresh object at each request.

    A new connection is created only when there are no connections left
    in the pool.
    """
    @staticmethod
    def instance():
        """
        Retrieve or construct a pool object.

        There is a global pool that is lazily constructed
        and returned by this method.
        """
        global curlPool
        # Reusing curl object after forking could cause problems
        # (e.g. shared sockets...), so we make sure that each
        # process has its own pool.
        pid = os.getpid()

        # There might be a race to create the pool, but
        # it probably isn't worth a lock. If two are created
        # the first one assigned to curlPool will just
        # lose a reference and be deleted - not a big deal.
        if curlPool is None or curlPool.pid != pid:
            curlPool = Pool(pycurl.Curl, pid)

        return curlPool

    def __init__(self, create, pid):
        """
        Initialize a pool.

        :param create: Function used to create new connections.
        :param pid: PID of the process creating the pool.
        """
        self.create = create
        self.pid = pid
        self.lock = threading.Lock()
        self.pool = []

    def get(self):
        """
        Retrieve a connection from the pool.

        The result should be returned after use (see :meth:`.put`).
        """
        self.lock.acquire()
        try:
            value = self.pool.pop()
        except IndexError:
            value = None
        finally:
            self.lock.release()

        # Create new ones outside the lock
        return value or self.create()

    def put(self, value):
        """
        Return a connection to the pool.
        """
        # We could call value.reset() here before returning the curl object
        # to the pool for pycurl version >= 7.19.0 if the C code for reset
        # actually incremented the reference on the returned None object.
        # As the C code is now, if called, the refcount on None eventually
        # goes to zero after enough requests and the Python interpreter
        # dies a quick death.
        # Note that this is fixed in pycurl 7.19.0.1 and patched in the RPM
        # package for pycurl 7.19.0 for CentOS 6.
        self.lock.acquire()
        try:
            self.pool.append(value)
        finally:
            self.lock.release()


def normalize_headers(headers):
    """
    Create a list of headers from:
       * A list of curl-style headers (return a copy)
       * None
       * a dictionary.

    The result is a list of curl-style headers.

    :param headers: List or dict of header (may also be None).
    :rtype headers: list[string] | dict[string, string] | None
    :return: A list of headers suitable for pycurl.
    :rtype: list[string]
    """
    if headers is None:
        result = []
    elif isinstance(headers, dict):
        result = ['%s: %s' % entry for entry in iteritems(headers)]
    else:
        result = headers[:]
    return [to_native_string(h) for h in result]


def retrying_perform(curl):
    """
    Perform a curl request, retrying in case of connection errors,

    We want to be able to resend the request if the current connection
    is unusable due to econnreset (e.g. because a proxy or a load balancer
    has just killed a keep-alive connection).

    :param curl: Curl object storing request parameters.
    :type curl: pycurl.Curl
    """
    # Delay before next retry (in seconds).
    retry = 0.1

    while True:
        try:
            curl.perform()
            break
        except pycurl.error as error:
            # Only retry in case of an ECONNRESET happening when
            # the connection is created.
            if (error.args[0] == pycurl.E_COULDNT_CONNECT and
                    curl.getinfo(pycurl.OS_ERRNO) == errno.ECONNRESET and
                    # Keep retrying until the delay reaches two seconds.
                    # This gives us 5 sleeps (and thus 6 requests,
                    # including the initial one) with the total sleep
                    # time of at most ~3.1 seconds.
                    retry < 2.0):
                # Retry
                time.sleep(retry)
                # The retry period will grow exponentially.
                retry *= 2
                continue
            # Not ECONNRESET or retry limit exceeded
            raise

# These constants are missing from pycurl <= 7.19.5, 
# some systems we want to support have only 7.19.0
# Note that libcurl itself must be at least 7.19.0 
# for this to work.
PROXYTYPE_SOCKS5_HOSTNAME = 7
PROXYTYPE_SOCKS4A = 6
PROXYTYPE_HTTP = 0

# Maps proxy type names to curl constants
# TODO: HTTPS proxies (since curl 7.52.0)?
_proxy_types = {
    'http': PROXYTYPE_HTTP,
    'socks': PROXYTYPE_SOCKS5_HOSTNAME,
    'socks4': PROXYTYPE_SOCKS4A,
    'socks5': PROXYTYPE_SOCKS5_HOSTNAME,
}


def makeRequest(obj, method, url, body=None, accept=None, contentType=None, callback=None, errCallback=None, headers=None):
    """
    Send a request to the server.

    See :func:`franz.miniclient.backends.requests.makeRequest` for documentation.
    """
    curl = Pool.instance().get()

    # TODO: Just make this an option
    # Uncomment these 5 lines to see pycurl debug output
    ## def report(debug_type, debug_msg):
    ##     if debug_type != 3:
    ##         print "debug(%d): %s" % (debug_type, debug_msg)
    ##curl.setopt(pycurl.VERBOSE, 1)
    ##curl.setopt(pycurl.DEBUGFUNCTION, report)

    #curl.setopt(pycurl.TIMEOUT, 45)

    if accept is None:
        accept = "*/*"

    # Proxy support
    if obj.proxy is not None:
        curl.setopt(pycurl.PROXY, obj.proxy_host)
        curl.setopt(pycurl.PROXYPORT, obj.proxy_port)
        curl.setopt(pycurl.PROXYTYPE, _proxy_types[obj.proxy_type])
    else:
        # Unsetopt doesn't work. As usual.
        curl.setopt(pycurl.PROXY, '')

    if obj.user is not None and obj.password is not None:
        curl.setopt(pycurl.USERPWD, "%s:%s" % (to_native_string(obj.user),
                                               to_native_string(obj.password)))
        curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
    else:
        curl.unsetopt(pycurl.USERPWD)
    if obj.cainfo is not None:
        curl.setopt(pycurl.CAINFO, copt(obj.cainfo))
    if obj.sslcert is not None:
        curl.setopt(pycurl.SSLCERT, copt(obj.sslcert))
    if obj.verifyhost is not None:
        curl.setopt(pycurl.SSL_VERIFYHOST, obj.verifyhost)
    if obj.verifypeer is not None:
        curl.setopt(pycurl.SSL_VERIFYPEER, obj.verifypeer)

    url = to_native_string(url)
    if not url.startswith("http:") and not url.startswith("https:"):
        url = to_native_string(obj.url) + to_native_string(to_native_string(url))

    method = to_native_string(method)

    # Usually POST and UPLOAD correspond to POST and PUT requests respectively.
    # In our case we will override the method using CUSTOMREQUEST and the
    # settings here will tell us if we're uploading using a READFUNCTION
    # or posting a string with POSTFIELDS.
    curl.setopt(pycurl.POST, 0)
    curl.setopt(pycurl.UPLOAD, 0)
    if body:
        if method in ("POST", "PUT"):
            if isinstance(body, basestring):
                # String
                body = to_bytes(body)
                curl.setopt(pycurl.POSTFIELDS, body)
                curl.setopt(pycurl.POST, 1)
            else:
                # File - can be passed as READDATA, but not as POSTFIELDS.
                curl.setopt(pycurl.READDATA, body)
                curl.setopt(pycurl.UPLOAD, 1)
        else:
            contentType = None
            url = url + "?" + to_native_string(body)
    else:
        contentType = None

    curl.setopt(pycurl.CUSTOMREQUEST, method)
    curl.setopt(pycurl.URL, url)

    # The "Expect:" is there to suppress "Expect: 100-continue"
    # behavior that is the default in libcurl when posting large
    # bodies.
    headers = normalize_headers(headers)
    if headers is None:
        headers = []
    headers.extend(["Connection: keep-alive", "Accept: " + to_native_string(accept), "Expect:"])
    if callback: headers.append("Connection: close")
    if contentType: headers.append("Content-Type: " + to_native_string(contentType))
    if obj.runAsName: headers.append("x-masquerade-as-user: " + to_native_string(obj.runAsName))
    curl.setopt(pycurl.HTTPHEADER, headers)
    curl.setopt(pycurl.ENCODING, "")  # which means 'any encoding that curl supports'

    if callback:
        status = [None]
        error = []

        # Called by curl for each header line.
        # The first line will contain the status code and that is
        # the only part we're interested in.
        def headerfunc(string):
            if status[0] is None:
                # Parse the status code if this is the first line.
                status[0] = int(unicode(string, 'utf-8').split(" ")[1])
            # return input length to indicate "no errors".
            return len(string)

        # Called by curl for each block of data received.
        # The argument will be a bytestring.
        def writefunc(string):
            if status[0] == 200: callback(string)
            else: error.append(unicode(string, 'utf-8'))

        curl.setopt(pycurl.WRITEFUNCTION, writefunc)
        curl.setopt(pycurl.HEADERFUNCTION, headerfunc)
        retrying_perform(curl)
        if status[0] != 200:
            errCallback(curl.getinfo(pycurl.RESPONSE_CODE), "".join(error))
    else:
        buf = io.BytesIO()
        curl.setopt(pycurl.WRITEFUNCTION, buf.write)
        retrying_perform(curl)
        response = to_native_string(buf.getvalue())
        buf.close()
        result = (curl.getinfo(pycurl.RESPONSE_CODE), response)
        Pool.instance().put(curl)
        return result



################################################################################
# Copyright (c) 2006-2017 Franz Inc.
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

# Just an alias, to avoid some circular imports without breaking
# backwards compatibility.
from franz.openrdf.query.repositoryresult import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals

from .rdfformat import RDFFormat
from .rdfwriter import RDFWriter

class RDFXMLWriter(RDFWriter):
    """
    An implementation of the RDFWriter interface that writes RDF documents in
    XML-serialized RDF format.
    This records the format as
    RDF/XML, and records the 'filePath' where the serialized RDF will
    be output to.  If 'filePath' is None, output is to standard output.
    """
    def __init__(self, filePath=None):
        super(RDFXMLWriter, self).__init__(RDFFormat.RDFXML, filePath)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103 

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

from .value import Value, BNode, URI
from .literal import Literal, CompoundLiteral, RangeLiteral, GeoCoordinate
from .statement import Statement

class ValueFactory(object):
    """
    A factory for creating URIs, blank nodes, literals and statements.
    """
    BLANK_NODE_AMOUNT = 10    
    def __init__(self, store):
        self.store = store
        self.unusedBNodeIds = []
        
    def getUnusedBNodeId(self):
        if not self.unusedBNodeIds:
            ## retrieve a set of bnode ids (they include leading '_:', which we strip off later:
            self.unusedBNodeIds = self.store.mini_repository.getBlankNodes(amount=ValueFactory.BLANK_NODE_AMOUNT)
        return self.unusedBNodeIds.pop()[2:] ## strip off leading '_:'

    def createBNode(self, nodeID=None):
        """
        See :meth:`.RepositoryConnection.createBNode`.
        """
        if not nodeID:
            nodeID = self.getUnusedBNodeId()
        return BNode(nodeID)
    
    def createLiteral(self, value, datatype=None, language=None):
        """
        See :meth:`.RepositoryConnection.createLiteral`.
        """
        if isinstance(value, (tuple, list)) and len(value) == 2:
            return self.createRange(value[0], value[1])

        return Literal(value, datatype=datatype, language=language)
        

    def createStatement(self, subject, predicate, _object, context=None):
        """
        See :meth:`.RepositoryConnection.createStatement`.
        """
        return Statement(subject, predicate, _object, context=context)
    
    def createURI(self, uri=None, namespace=None, localname=None, canonical=True):
        """
        See :meth:`.RepositoryConnection.createURI`.
        """
        if namespace is not None and localname is None:
            return URI(namespace=uri, localname=namespace, canonical=canonical)
        else:
            return URI(uri=uri, namespace=namespace, localname=localname, canonical=canonical)


#############################################################################
## Extension to RDF4J API
#############################################################################

    def validateRangeConstant(self, term, predicate):
        """Validate an individual range constant"""
        datatype = term.getDatatype()
        if not datatype:
            raise Exception('Illegal term in range expression "%s" needs to '
                            'have a datatype.' % term.getValue())

    def validateCompoundLiteral(self, term, predicate):
        """
        Check to see if range boundaries are mapped.
        TODO: ADD VALIDATION FOR GEO TERMS
        """
        if isinstance(term, RangeLiteral):
            self.validateRangeConstant(term.lowerBound, predicate)
            self.validateRangeConstant(term.upperBound, predicate)
        elif isinstance(term, GeoCoordinate):
            pass
        
    def object_position_term_to_openrdf_term(self, term, predicate=None):
        """
        If 'term' is a string, integer, float, etc, convert it to
        a Literal term.  Otherwise, if its a Value, just pass it through.
        """
        if term is not None:
            if isinstance(term, CompoundLiteral): 
                self.validateCompoundLiteral(term, predicate)
            elif not isinstance(term, Value):
                term = self.createLiteral(term)
        return term
    
    def createRange(self, lowerBound, upperBound):
        """
        Create a compound literal representing a range from 'lowerBound' to 'upperBound'
        """
        lowerBound = self.object_position_term_to_openrdf_term(lowerBound)
        upperBound = self.object_position_term_to_openrdf_term(upperBound)
        return RangeLiteral(lowerBound=lowerBound, upperBound=upperBound)

    def namespace(self, prefix, canonical=True):
        """
        Create an object that allows for simple creation of URIs in given namespace.
        Attribute lookups on the returned object will produce URIs with the attribute
        name as localname. Array item access and function calls will have the same
        effect.

        :param prefix: Prefix prepended to URIs created by the returned object.
        :type prefix: str
        :param canonical: If true (default) ensure that the same URI object
                          is returned each time when the same string is
                          passed as the local name.
        :type canonical: bool
        :return: An object that can be used to create URIs.
        """
        vf = self

        class Namespace(object):
            __slots__ = ()

            def __getattribute__(self, name):
                return self(name)

            def __getitem__(self, index):
                return self(index)

            def __call__(self, arg):
                return vf.createURI(namespace=prefix, localname=arg, canonical=canonical)

        return Namespace()

    def namespaces(self, prefixes, canonical=True):
        """
        Same as :meth:`.namespace`, but allows to create URIs in multiple namespaces
        from the strings of the form '<abbrev>:<name>', where <abbrev> is one of
        the namespace abbreviations provided in the 'prefixes' dictionary, and
        <name> is the resource name.

        :param prefixes: Dictionary mapping prefix abbreviations to prefixes to
                         be prepended to URIs created by the returned object.
        :type prefixes: dict
        :param canonical: If true (default) ensure that the same URI object
                          is returned each time when the same string is
                          passed as the local name.
        :type canonical: bool
        :return: An object that can be used to create URIs.

        """
        vf = self
        pdict = prefixes.copy()

        class Namespaces(object):
            __slots__ = ()

            def __getattribute__(self, name):
                return self(name)

            def __getitem__(self, index):
                return self(index)

            def __call__(self, arg):
                components = arg.split(':')
                abbrev, name = components[0], ':'.join(components[1:])
                prefix = pdict.get(abbrev)
                assert prefix is not None, "Unknown namespace abbreviation: %s" % abbrev
                return vf.createURI(namespace=prefix, localname=name, canonical=canonical)

        return Namespaces()



# pylint: disable-msg=C0111

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals

from .literal import Literal
from .statement import Statement
from .value import Value, URI, BNode
from .valuefactory import ValueFactory
from .utils import parse_term

__all__ = ['BNode', 'Literal', 'Statement', 'URI',
    'Value', 'ValueFactory', 'parse_term']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from future.builtins import object
from past.builtins import basestring

from future.utils import iteritems, python_2_unicode_compatible

from franz.openrdf.rio.rdfformat import RDFFormat
from franz.openrdf.rio.tupleformat import TupleFormat
from franz.openrdf.util.contexts import output_to
from ..exceptions import IllegalOptionException, QueryMissingFeatureException
from .dataset import ALL_CONTEXTS, Dataset
from .queryresult import GraphQueryResult, TupleQueryResult


@python_2_unicode_compatible
class QueryLanguage(object):
    registered_languages = []
    SPARQL = None
    PROLOG = None

    def __init__(self, name):
        self.name = name
        QueryLanguage.registered_languages.append(self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'QueryLanguage.' + self.name

    def getName(self):
        return self.name

    @staticmethod
    def values(): return QueryLanguage.registered_languages[:]
    
    @staticmethod
    def valueOf(name): 
        for ql in QueryLanguage.registered_languages:
            if ql.name.lower() == name.lower():
                return ql
        return None
    
QueryLanguage.SPARQL = QueryLanguage('SPARQL')
QueryLanguage.PROLOG = QueryLanguage('PROLOG')

#############################################################################
##
#############################################################################

class Query(object):
    """
    A query on a repository that can be formulated in one of the
    supported query languages (for example SeRQL or SPARQL). It allows one to
    predefine bindings in the query to be able to reuse the same query with
    different bindings.
    """
    def __init__(self, queryLanguage=QueryLanguage.SPARQL,
                 query=None, baseURI=None, queryString=None):
        """
        Initialize a query object.

        :param queryLanguage: Query syntax - the default is SPARQL.
        :type queryLanguage: QueryLanguage
        :param query: Query text
        :type query: string
        :param baseURI: Prefix used to relsove relative URIs.
                        Default: chosen by the server.
        :type baseURI: URI
        :param queryString: Legacy name of the 'query' parameter.
        :type queryString: string
        """
        if query is None:
            raise TypeError('query argument is missing.')
        self.queryLanguage = Query._check_language(queryLanguage)
        self.queryString = query or queryString
        self.baseURI = baseURI
        self.dataset = None
        self.includeInferred = False
        self.bindings = {}
        self.connection = None
        self.checkVariables = False
        ## CommonLogic parameters:
        self.preferred_execution_language = None
        self.actual_execution_language = None
        self.subject_comes_first = False

    def setBinding(self, name, value):
        """
        Binds the specified variable to the supplied value. Any value that was
        previously bound to the specified value will be overwritten.

        :param name: Variable name.
        :type name: string
        :param value: New value for the variable. If a string is passed it will
                      be converted to a literal.
        :type value: Value|basestring
        """
        if isinstance(value, basestring):
            value = self._get_connection().createLiteral(value)
        self.bindings[name] = value
        
    def setBindings(self, dict):
        """
        Set multiple variable bindings.

        :param dict: A dictionary of bindings.
        :type dict: dict[string, Value]
        """
        if not dict: return
        for key, value in iteritems(dict):
            self.setBinding(key, value)

    def removeBinding(self, name):
        """ 
        Removes a previously set binding on the supplied variable. Calling this
        method with an unbound variable name has no effect.

        :param name: Variable name.
        :type name: string
        """
        self.bindings.pop(name)

    def getBindings(self):
        """
        Retrieve the bindings that have been set on this query.

        :return: A dictionary of bindings.
        :rtype: dict[string, Value]
        """ 
        return self.bindings

    def setDataset(self, dataset):
        """
        Select the dataset against which to evaluate this query, overriding any
        dataset that is specified in the query itself.

        :param dataset: A dataset object.
        :type dataset: Dataset
        """ 
        self.dataset = dataset
     
    def getDataset(self):
        """
        Get the dataset against which this query will operate.

        :return: Dataset.
        :rtype: Dataset
        """ 
        return self.dataset
    
    def setContexts(self, contexts):
        """
        Assert a set of contexts (named graphs) that filter all triples.

        :param contexts: List of graph URIs.
        :type contexts: list[URI|string]
        """
        if not contexts: return
        ds = Dataset()
        for cxt in contexts:
            if isinstance(cxt, basestring): cxt = self._get_connection().createURI(cxt)
            ds.addNamedGraph(cxt)
        self.setDataset(ds)
     
    def setIncludeInferred(self, includeInferred):
        """
        Determine whether evaluation results of this query should include inferred
        statements.

        The default setting is ``True``.

        :param includeInferred: If ``True`` include inferred triples in query results.
        :type includeInferred: bool
        """
        self.includeInferred = includeInferred

    def getIncludeInferred(self):
        """
        Check whether this query will return inferred statements.
        :return: ``True`` if inferred triples are included, ``False`` otherwise.
        :rtype: bool
        """ 
        return self.includeInferred
    
    def setCheckVariables(self, setting):
        """
        Determine whether the presence of variables in the select clause not referenced
        in a triple is flagged.

        :param setting: ``True`` if variables should be checked, ``False`` otherwise.
        :type setting: bool
        """
        self.checkVariables = setting
    
    def setConnection(self, connection):
        """
        Internal call to embed the connection into the query.
        """
        self.connection = connection
        
    def _get_connection(self):
        return self.connection
    
    def evaluate_generic_query(self, count=False, accept=None, callback=None,
                               analyze=False, analysisTechnique=None, analysisTimeout=None,
                               update=False):
        """
        Evaluate a SPARQL or PROLOG query, which may be a 'select', 'construct', 'describe'
        or 'ask' query (in the SPARQL case). Return an appropriate response.

        :param count: If ``True`` return only the count of the query results.
        :type count: bool
        :param accept: Set to ``'application/sparql-results+xml'`` or ``'application/sparql-results+json'``
                       to retrieve the result as a string in the specified format.
                       The default is to return a :class:`QueryResult` object.
        :type accept: string
        :param analyze: If ``True`` perform query analysis for SPARQL queries.
        :type analyze: bool
        :param analysisTechnique: Control the method of analysis. Can be either `"executed"` (default).
                                  meaning that the query will be executed to perform dynamic analysis,
                                  or ``"static"``.
        :type analysisTechnique: string
        :param analysisTimeout: Number of second to run the query if executed for analysis.
        :type analysisTimeout: float
        :param update: If ``True`` this is an update query and the result will be a boolean.
        :type update: bool
        :return: The result can be either:
                    - A boolean (for update queries).
                    - A dictionary containing the results.
                    - A string (if ``accept`` was passed).
                    - An integer (if ``count=True``).
        :rtype: dict|string|int|bool
        """
        conn = self._get_connection()
        namedContexts = conn._contexts_to_ntriple_contexts(
                        self.dataset.getNamedGraphs() if self.dataset else None)
        regularContexts = conn._contexts_to_ntriple_contexts(
                self.dataset.getDefaultGraphs() if self.dataset else ALL_CONTEXTS)
        bindings = None
        if self.bindings:
            bindings = {}
            for vbl, val in list(self.bindings.items()):
                bindings[vbl] = conn._convert_term_to_mini_term(val)
        mini = conn._get_mini_repository()
        if self.queryLanguage == QueryLanguage.SPARQL:  
            response = mini.evalSparqlQuery(self.queryString, context=regularContexts, namedContext=namedContexts, 
                                            infer=self.includeInferred, bindings=bindings,
                                            checkVariables=self.checkVariables, count=count,
                                            accept=accept, callback=callback,
                                            analyze=analyze, analysisTechnique=analysisTechnique,
                                            analysisTimeout=analysisTimeout, update=update)
        elif self.queryLanguage == QueryLanguage.PROLOG:
            if namedContexts:
                raise QueryMissingFeatureException("Prolog queries do not support the datasets (named graphs) option.")
            if analyze:
                raise QueryMissingFeatureException("Prolog queries do not support analysis.")
            # evalPrologQuery is already always done as if update=True
            response = mini.evalPrologQuery(self.queryString,
                                            infer=self.includeInferred, count=count, accept=accept,
                                            callback=callback)
        else:
            raise ValueError('Unsupported query language: %s' % self.queryLanguage)
        return response

    @staticmethod
    def _check_language(queryLanguage):
        if queryLanguage == 'SPARQL':
            return QueryLanguage.SPARQL

        if queryLanguage == 'PROLOG':
            return QueryLanguage.PROLOG

        if not queryLanguage in [QueryLanguage.SPARQL, QueryLanguage.PROLOG]:
            raise IllegalOptionException("Can't evaluate the query language '%s'.  Options are: SPARQL and PROLOG."
                                         % queryLanguage)
        return queryLanguage
       
  
class TupleQuery(Query):
    """
    A query that returns tuples (i.e. sets of variable bindings).
    """
    def evaluate(self, count=False, output=None,
                 output_format=TupleFormat.TABLE):
        """Execute the embedded query against the RDF store.

        Return an iterator that produces for each step a tuple of values
        (resources and literals) corresponding to the variables or
        expressions in a 'select' clause (or its equivalent).

        :param count: If ``True`` return the number of result rows
                      instead of the usual iterator.
        :type count: bool
        :param output: A file name, file descriptor or a file-like
                       object to save the results to
                       (optional). ``True`` can be used as a synonym
                       for stdout.
        :type output: str|file|int|bool
        :param output_format: Serialization format for ``output``.
                              The default is TABLE.
        :type output_format: RDFFormat
        :return: Either an iterator over results, 
                 an integer (the number of results, if ``count`` is used) 
                 or ``None`` (if ``output`` is used).
        :rtype: TupleQueryResult|int
        """

        with output_to(output) as out:
            callback = None if output is None else out.write
            accept = None if output is None else TupleFormat.mime_type_for_format(output_format)
            response = self.evaluate_generic_query(
                count=count, accept=accept, callback=callback)

            if count:
                return response

            if output is not None:
                return None

            return TupleQueryResult(response['names'], response['values'], response.get('queryInfo'))

    def analyze(self, analysisTechnique=None, analysisTimeout=None):
        """
        Analyze the query.

        Analysis is only available for SPARQL queries.

        :param analysisTechnique: Control the method of analysis. Can be either `"executed"` (default).
                                  meaning that the query will be executed to perform dynamic analysis,
                                  or ``"static"``.
        :type analysisTechnique: string
        :param analysisTimeout: Number of second to run the query if executed for analysis.
        :type analysisTimeout: float
        :return: Analysis result as a dictionary.
        :rtype: dict
        """
        response = self.evaluate_generic_query(analyze=True, analysisTechnique=analysisTechnique,
            analysisTimeout=analysisTimeout)
        return response


class UpdateQuery(Query):
    """
    An update query.
    """
    def evaluate(self):
        """
        Execute the embedded update against the RDF store.

        :return: ``True`` if the query changed the store, ``False`` otherwise.
        :rtype: bool
        """
        return self.evaluate_generic_query(update=True)

class GraphQuery(Query):
    """
    A query that returns statements (see :class:`Statement`).
    """
    def evaluate(self, output=None, output_format=RDFFormat.TABLE):
        """
        Execute the embedded query against the RDF store.

        :param output: A file name, file descriptor or a file-like
                       object to save the results to
                       (optional). ``True`` can be used as a synonym
                       for stdout.
        :type output: str|file|int|bool
        :param output_format: Serialization format for ``output``.
                              The default is TABLE.
        :type output_format: RDFFormat
        :return: An iterator over statements or None (if ``output`` is used)..
        :rtype: GraphQueryResult
        """
        with output_to(output) as out:
            callback = None if output is None else out.write
            accept = None if output is None else RDFFormat.mime_type_for_format(output_format)
            response = self.evaluate_generic_query(accept=accept, callback=callback)
            if output is None:
                return GraphQueryResult(response)
            else:
                return None


class BooleanQuery(Query):
    """
    A query that returns a boolean.
    """
    def evaluate(self):
        """
        Execute the embedded query against the RDF store.

        :return: A boolean.
        :rtype: bool
        """
        return self.evaluate_generic_query()



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

"""
Test fixtures, moved to a separate file to avoid warnings
about parameter names shadowing global symbols.
"""
import random
from collections import MutableMapping
import uuid

import pytest
import requests

from franz.openrdf.model import URI
from franz.openrdf.repository import Repository
from franz.openrdf.repository.attributes import AttributeDefinition
from franz.openrdf.sail import AllegroGraphServer

from .tests import AG_HOST, AG_PORT, AG_PROXY, STORE, CATALOG, USER, PASSWORD
import contextlib, os, threading
import six
from six.moves import BaseHTTPServer


# noinspection PyShadowingNames
@pytest.fixture
def conn(server):
    """
    Provides a connection to the test repository. The repository is cleared
    before the test, but not after it.
    """
    catalog = server.openCatalog(CATALOG)
    stores = catalog.listRepositories()

    # Instead of renewing the database, clear it.
    mode = Repository.CREATE if STORE not in stores else Repository.OPEN

    repo = catalog.getRepository(STORE, mode)
    repo.initialize()
    connection = repo.getConnection()

    if STORE in stores:
        connection.setUserAttributes({})
        connection.clearAttributeFilter()
        connection.clear()
        connection.clearNamespaces()
        connection.disableDuplicateSuppression()
        for attr in connection.getAttributeDefinitions():
            connection.deleteAttributeDefinition(attr.name)

    with connection:
        yield connection


# noinspection PyShadowingNames
@pytest.fixture
def repo_name(conn):
    """
    Make sure that a repository exists and is empty.

    Return the name of that repository.
    """
    name = conn.repository.database_name
    conn.close()
    return name


@pytest.fixture
def server():
    """
    A fixture that allows easy access to an AllegroGraphServer object.
    """
    yield AllegroGraphServer(AG_HOST, AG_PORT, USER, PASSWORD, proxy=AG_PROXY)


# noinspection PyShadowingNames
@pytest.fixture
def non_existing_repo(server):
    """
    A fixture that makes sure a repository does not exist before the test
    and also deletes it after the test.
    The value returned by this fixture is the name of the store.
    """
    catalog = server.openCatalog(CATALOG)
    store = 'temp-store'
    while store in catalog.listRepositories():
        store = 'temp-store-' + ''.join(random.choice('0123456789') for _ in range(16))
    yield store
    if store in catalog.listRepositories():
        catalog.deleteRepository(store)


# noinspection PyShadowingNames
@pytest.fixture
def session(conn):
    conn.openSession()
    yield conn
    conn.closeSession()


@pytest.fixture
def attr(conn):
    attr = AttributeDefinition(
        name='test', allowed_values=['a', 'b', 'c'],
        minimum_number=0, maximum_number=1)
    yield conn.setAttributeDefinition(attr)


@pytest.fixture
def attr2(conn):
    attr = AttributeDefinition(
        name='test2', allowed_values=['a', 'b', 'c'],
        minimum_number=0, maximum_number=None)
    yield conn.setAttributeDefinition(attr)


@pytest.fixture
def clean_env():
    old = os.environ
    os.environ = {}
    yield os.environ
    os.environ = old


# Injecting common URIs this way is less verbose


def make_uri_fixture(name):
    """
    Create a pytest fixture named NAME that resolves
    to a URI object '<ex://NAME>'.
    """
    # noinspection PyShadowingNames
    def func(conn):
        return conn.createURI('ex://' + name)
    func.__name__ = name
    return pytest.fixture(func, name=name)


def define_uri_fixtures():
    gl = globals()
    for uri in ('s', 'p', 'o', 'g', 'x'):
        for suffix in ('', '1', '2'):
            name = uri + suffix
            gl[name] = make_uri_fixture(name)


define_uri_fixtures()


@pytest.fixture
def ex(conn):
    return conn.namespace('ex://')


@pytest.fixture
def example(conn):
    return conn.namespace('http://franz.com/example/')


class UserData(MutableMapping):
    """
    A dictionary-like object that stores values in the user data
    service on an AllegroGraph server.

    This object keeps track of the values it changed. All changes
    can be undone by calling 'close'.
    """
    def __init__(self, server_object):
        """
        Initialize user data dictionary.

        :param server_object: AG server handle.
        :type server_object: AllegroGraphServer
        """
        self.server = server_object
        self.old_values = {}

    def __setitem__(self, k, v):
        if k not in self.old_values:
            self.old_values[k] = self.server.getUserData(k)
        self.server.setUserData(k, v)

    def __delitem__(self, k):
        if k not in self.old_values:
            self.old_values[k] = self.server.getUserData(k)
        self.server.deleteUserData(k)

    def __getitem__(self, k):
        return self.server.getUserData(k)

    def __len__(self):
        return len(self.server.listUserData())

    def __iter__(self):
        for k in self.server.listUserData():
            yield self.server.getUserData(k)

    def close(self):
        """
        Undo all changes made through this object.
        """
        for k, v in six.iteritems(self.old_values):
            if v is None:
                self.server.deleteUserData(k)
            else:
                self.server.setUserData(k, v)
        self.old_values.clear()


# noinspection PyShadowingNames
@pytest.fixture
def user_data(server):
    """
    A test fixture providing access to an instance of the UserData class.
    All user data changes are undone during cleanup.
    """
    with contextlib.closing(UserData(server)) as ud:
        yield ud


class HTTPServer(object):
    """
    A trivial HTTP server used during tests.

    After creation handlers for specific paths can be added with `add_handler()`.
    Calling `start()` will cause the server to listen for connections in
    a separate thread, calling `close()` will stop the server.

    The server only handles GET requests.

    This is very basic, features should be added as needed by tests.
    """
    def __init__(self):
        class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
            def do_DELETE(self):
                self.send_response(200, 'OK')
                self.send_header('Content-Type', 'text/plain')
                self.send_header('Content-Length', '0')
                self.end_headers()
                server_self.keep_running = False

            def do_GET(self):
                path = ensure_slash(self.path)

                if path not in handlers:
                    self.send_response(404, 'Not found')
                    self.send_header('Content-Type', 'text/plain')
                    self.end_headers()
                    self.wfile.write(b'Oops: ' + path.encode('utf-8'))
                else:
                    self.send_response(200, 'OK')
                    self.send_header('Content-Type', 'text/plain')
                    self.end_headers()
                    handler = handlers[self.path]
                    result = handler() if callable(handler) else handler
                    self.wfile.write(result)

        server_self = self
        self.handlers = handlers = {}
        self.server = BaseHTTPServer.HTTPServer(('127.0.0.1', 0), Handler)
        self.keep_running = True
        self.thread = threading.Thread(target=self._run)
        # Do not hang if a server thread refuses to die
        self.thread.daemon = True

    def _run(self):
        while self.keep_running:
            self.server.handle_request()
        self.server.server_close()

    def start(self):
        """
        Start listening for connections.
        """
        self.thread.start()

    def close(self):
        """
        Stop the server.
        """
        requests.delete(self.url('anything'))
        self.thread.join(5.0)
        if self.thread.is_alive():
            raise Exception('Embedded HTTP server refused to terminate!')

    def publish(self, path, handler):
        """
        Add a handler for a specific path.

        :param path: Path to be handled.
        :param handler: Content to return, either a bytestring or a function
                        that returns one.
        """
        path = ensure_slash(path)
        self.handlers[path] = handler

    def url(self, path):
        """
        Get the full URL for a specific path.

        :param path: Path to be accessed.
        :return: A full URL, including the port number.
        """
        path = ensure_slash(path)
        return 'http://127.0.0.1:%d%s' % (self.server.server_port, path)


@pytest.fixture
def http_server():
    """
    A fixture wrapping the HTTP server, takes care of starting and stopping it.
    :return:
    """
    with contextlib.closing(HTTPServer()) as server:
        server.start()
        yield server


# Lisp code used to run an HTTP server.
# Note that it is a .format() string and expects a security key
# to be passed in the {key} keyword argument.
server_code = """
(use-package :net.aserve)

(defun fixed-handler (body)
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (write-string body (request-reply-stream req))))))

(defun pub (req ent)
  (when (string= (header-slot-value req :authorization) "test {key}")
    (with-http-response (req ent)
      (with-http-body (req ent)
        (let ((body (get-request-body req))
              (path (request-query-value "path" req))
              (type (or (request-query-value "type" req) "text/plain")))
          (publish :path path
                   :content-type type 
                   :function (fixed-handler body)
                   :server (request-wserver req)))))))
  
(defun stop (req ent)
  (with-http-response (req ent)
    (with-http-body (req ent)))
  (when (string= (header-slot-value req :authorization) "test {key}")
    (shutdown :server (request-wserver req))))
  
(let ((server (start :port 0 :host "127.0.0.1" :server :new)))
  (publish :path "/__pub"
           :content-type "text/plain"
           :server server
           :function #'pub)
  (publish :path "/__stop" :function #'stop :server server)
  (socket:local-port (wserver-socket server)))
"""


def lisp_string(python_string):
    """
    Convert a string to a Lisp string literal.
    """
    return '"%s"' % python_string.replace('\\', '\\\\').replace('"', '\\"')


class RemoteHTTPServer(object):
    """
    Used to make things available through HTTP from the AG server.

    The server is started by calling `start()` and stopped by
    calling `close()`. New documents (strings) can be published
    by using `publish()`. Full URIs to documents (including
    the port number) can be obtained by calling `uri()`.

    Note that documents published here are not necessarily accessible
    from the machine on which the tests are run. Any requests for such
    data must be executed on the AG server. The `send_request()` method
    can be used to send a request through the AG server and return
    the response.

    Very basic, works only with text, can only serve simple strings.
    Should be extended when needed.
    """
    def __init__(self, connection):
        self.conn = connection
        self.key = uuid.uuid4()
        self.port = 0  # Will be set later

    def send_request(self, path, body=None, params=None):
        """
        Execute a HTTP request *from the AG server*.

        This will be a GET request, unless the `body` parameter is used,
        in which case a POST request will be sent.

        :param path: Document path (as given to `publish()`.
        :param body: Request body - if specified this will be a POST request.
        :param params: Request parameters, passed in the URI.
        :return: Response body.
        """
        alist = None
        if params:
            pairs = ' '.join('(%s . %s)' % (lisp_string(k), lisp_string(v))
                             for k, v in six.iteritems(params))
            alist = "'(%s)" % pairs

        uri = self.uri(path)

        return self.conn.evalInServer('''
            (prog1 (net.aserve.client:do-http-request 
             {uri} 
             :method {method}
             :query {query}
             :headers '((:authorization . "test {key}"))
             :content-type "text/plain"
             :content {body}))
        '''.format(uri=lisp_string(uri),
                   query=alist or 'nil',
                   key=self.key,
                   method=':post' if body else ':get',
                   body=lisp_string(body) if body else 'nil'))

    def uri(self, path):
        """
        Retrieve full URI to a document published with `publish`.

        Note that the resulting URI is not necessarily accessible
        from the machine on which the tests are executed.

        :param path: Document path.
        :return: Full URI.
        """
        path = ensure_slash(path)
        return 'http://127.0.0.1:%d%s' % (self.port, path)

    def start(self):
        """
        Start listening for connections.
        """
        self.port = self.conn.evalInServer(server_code.format(key=self.key))

    def close(self):
        """
        Shutdown the server.
        """
        self.send_request('__stop')

    def publish(self, path, document, content_type='text/plain'):
        """
        Publish a document (string).

        :param path: Target path on the server.
        :param document: Document text.
        :param content_type: MIME type, the default is 'text/plain'.
        :return: Full URI to the published document.
        """
        path = ensure_slash(path)
        self.send_request('__pub', document, {
            'path': path,
            'type': content_type
        })
        return self.uri(path)


# noinspection PyShadowingNames
@pytest.fixture
def remote_http_server(conn):
    """
    A fixture wrapping the remote (i.e. AG-side) HTTP server,
    takes care of starting and stopping the server.
    """
    with contextlib.closing(RemoteHTTPServer(conn)) as server:
        server.start()
        yield server


def ensure_slash(text):
    """
    Prepend a slash to a string, unless it starts with one already.
    """
    if text.startswith('/'):
        return text
    return '/' + text


# Cache the version number
_version = None


def get_version():
    global _version
    if _version is None:
        server = AllegroGraphServer(AG_HOST, AG_PORT, USER, PASSWORD, proxy=AG_PROXY)
        _version = server.versionTuple
    return _version

@pytest.fixture(scope='module')
def version():
    return get_version()


# Decorator used to skip tests
def min_version(*args):
    version = get_version()
    if version < args:
        expected = '.'.join(str(c) for c in args)
        actual = '.'.join(str(c) for c in version)
        return pytest.mark.skip("AG server version: %s < %s" % (actual, expected))
    return lambda x: x


# noinspection PyShadowingNames
@pytest.fixture
def ex(conn):
    return conn.namespace('ex://')


@pytest.fixture
def after():
    """
    A fixture that resolves to a mutable list of functions that will
    be called after the test.
    """
    to_call = []
    exceptions = []
    yield to_call
    for o in to_call:
        try:
            o()
        except Exception as e:
            exceptions.append(e)
    if exceptions:

        raise Exception('Exception(s) raised when cleaning up' +
                        ('\n'.join(str(e) for e in exceptions)))



################################################################################
# Copyright (c) 2006-2017 Franz Inc.
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

"""
Utilities related to context managers.
"""
from contextlib import contextmanager
import doctest
import io
import os
import sys

from past.builtins import basestring


@contextmanager
def wrap_context(value):
    """
    Wrap a no-op context around an arbitrary value, so that it can be
    used in a ``with`` statement.

    :param value: Value to be returned by the context manager.
    :return: A context manager wrapping ``value``.
    """
    yield value


@contextmanager
def output_to(target):
    """
    Returns a file-like object constructed from target.

    Target can be:

        - None: the returned stream will ignore all data
        - True: data will be written to stdout.
        - a string: a file will be opened (in binary mode)
        - a file-like object: will be returned as is.
        - an int - a file descriptor.

    :param target: File path, None or a file-like object.
    """
    if target is None:
        # TODO: consider replacing with an object with empty methods
        target = os.devnull

    # Note that True is also an int (and it equals 1 = stdout),
    # but we give it a special treatment because our doctests
    # are not smart enough to capture things written to an fd.
    if target is True:
        # Make sure that the thing we return supports binary I/O
        if hasattr(sys.stdout, "buffer"):
            yield sys.stdout.buffer
        # Doctest checker only accepts strings, not bytes...
        elif isinstance(sys.stdout, doctest._SpoofOut):
            yield BinWrapper(sys.stdout)
        # io.TextIOBase only accepts strings, not bytes...
        elif isinstance(sys.stdout, io.TextIOBase):
            yield BinWrapper(sys.stdout)
        else:
            yield sys.stdout
    elif isinstance(target, int):
        with os.fdopen(target, "wb") as out:
            yield out
    elif isinstance(target, basestring):
        with open(target, "wb") as out:
            yield out
    elif hasattr(target, "write"):
        # If it quacks like a duck and has a write method...
        yield target
    else:
        raise ValueError("Invalid output specification %s" % target)


# Used to convert text output streams to binary UTF-8 streams
class BinWrapper(object):
    def __init__(self, target):
        self.target = target

    def write(self, byte_string):
        self.target.write(byte_string.decode("utf-8"))



import pandas
from six.moves import zip
from collections import OrderedDict

from franz.openrdf.model import Literal


def term_to_pandas(term):
    if term is None:
        return None
    if isinstance(term, Literal):
        return term.toPython()
    else:
        # URI or blank node
        return str(term)


def rows_to_pandas(rows, column_names):
    columns = [[] for _ in column_names]
    for row in rows:
        for idx, term in enumerate(row):
            columns[idx].append(term_to_pandas(term))
    return pandas.DataFrame(OrderedDict(zip(column_names, columns)))



# -*- coding: utf-8 -*-

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################
import json
from datetime import datetime, date, time, timedelta
from decimal import Decimal

import pytest
import requests
import io
import os
import sys

import re
from six import BytesIO

from future.utils import iteritems, iterkeys
from nose.tools import assert_raises

from franz.openrdf.connect import ag_connect
from franz.openrdf.exceptions import RequestError
from franz.openrdf.model import Literal, Statement, URI
from franz.openrdf.query.query import QueryLanguage
from franz.openrdf.repository.attributes import AttributeDefinition, TripleAttribute, UserAttribute, \
    attribute_filter_to_expr, And, Or, Not, Empty, Overlap, Subset, Superset, Equal, Lt, Le, Eq, Ge, Gt, \
    attribute_set_to_expr
from franz.openrdf.repository.transactions import TransactionSettings
from franz.openrdf.rio.docformat import DocFormat
from franz.openrdf.rio.rdfformat import RDFFormat
from franz.openrdf.rio.tupleformat import TupleFormat
from franz.openrdf.sail import AllegroGraphServer
from franz.openrdf.tests.conftest import min_version
from franz.openrdf.tests.tz import MockTimezone
from franz.openrdf.util.contexts import output_to
from franz.openrdf.util.http import normalize_headers
from franz.openrdf.vocabulary import XMLSchema

from franz.openrdf.tests.tests import AG_HOST, AG_PORT, AG_PROXY, CATALOG, STORE, USER, PASSWORD

# Imported to allow mocking
import franz.miniclient.request

common_args = dict(
    host=AG_HOST, port=AG_PORT, catalog=CATALOG,
    user=USER, password=PASSWORD, proxy=AG_PROXY
)


def get_statements(conn):
    """
    Gets all statements as a list of 4-element lists of Value objects,
    sorted lexicographically.
    """
    return sorted([list(s) for s in conn.getStatements()])


def normalize_query_result(result, sort=True):
    """
    Post-process query result to generate a simple, nested list.

    :param result: A QueryResult object.
    :param sort: if True (default) rows will be sorted.
    :return: A list of lists of RDF values.
    """
    normalized = [[row[i] for i in range(len(row))] for row in result]
    return sorted(normalized) if sort else normalized


def test_ag_connect_open(repo_name):
    with ag_connect(repo_name, create=False, **common_args) as conn:
        assert conn.size() == 0


def test_ag_connect_create(non_existing_repo):
    with ag_connect(non_existing_repo, create=True, **common_args) as conn:
        assert conn.size() == 0


def test_ag_connect_recreate(conn):
    store = conn.repository.database_name
    with conn:
        conn.addTriple('<http://franz.com/s>', '<http://franz.com/p>', '<http://franz.com/o>')
    with ag_connect(store, clear=True, **common_args) as conn:
        assert conn.size() == 0


def test_ag_connect_open_no_create(non_existing_repo):
    with pytest.raises(Exception):
        ag_connect(non_existing_repo, create=False, **common_args)


def test_ag_connect_create_exists():
    with pytest.raises(Exception):
        ag_connect(STORE, create=True, fail_if_exists=True, **common_args)


def test_ag_connect_fail_if_exists_but_not_create(repo_name):
    with ag_connect(repo_name, create=False, fail_if_exists=True, **common_args) as conn:
        assert conn.size() == 0


def test_ag_connect_session(repo_name):
    with ag_connect(repo_name, create=False, session=True, **common_args) as conn:
        assert conn.is_session_active


def test_server_all_data_in_host(clean_env):
    server = AllegroGraphServer('https://somehost:4321/and/then/some')
    assert server.url == 'https://somehost:4321/and/then/some'


def test_server_all_defaults(clean_env):
    server = AllegroGraphServer('somehost')
    assert server.url == 'http://somehost:10035'


def test_server_protocol_as_arg(clean_env):
    server = AllegroGraphServer('somehost', protocol='https')
    assert server.url == 'https://somehost:10036'


def test_server_override_protocol(clean_env):
    server = AllegroGraphServer('http://somehost', protocol='https')
    assert server.url == 'https://somehost:10036'


def test_server_port_as_arg(clean_env):
    server = AllegroGraphServer('somehost', port=4321)
    assert server.url == 'http://somehost:4321'


def test_server_override_port(clean_env):
    server = AllegroGraphServer('somehost:1234', port=4321)
    assert server.url == 'http://somehost:4321'


def test_server_https_if_cainfo(clean_env):
    server = AllegroGraphServer('somehost', cainfo='/path/to/ca/bundle')
    assert server.url == 'https://somehost:10036'


def test_server_url_in_env(clean_env):
    clean_env['AGRAPH_HOST'] = 'somehost'
    clean_env['AGRAPH_PORT'] = 12345
    clean_env['AGRAPH_USER'] = 'luser'
    clean_env['AGRAPH_PASSWORD'] = '1234'
    server = AllegroGraphServer()
    assert server.url == 'http://somehost:12345'
    assert server._client.user == 'luser'
    assert server._client.password == '1234'


@pytest.mark.parametrize("filename, expected_format, expected_compression", [
    ("somefile.rdf", "RDF/XML", None),
    ("somefile.rdfs", "RDF/XML", None),
    ("somefile.owl", "RDF/XML", None),
    ("somefile.xml", "RDF/XML", None),
    ("somefile.rdf.gz", "RDF/XML", "gzip"),
    ("somefile.rdfs.gz", "RDF/XML", "gzip"),
    ("somefile.owl.gz", "RDF/XML", "gzip"),
    ("somefile.xml.gz", "RDF/XML", "gzip"),
    ("somefile.nt", "N-Triples", None),
    ("somefile.ntriples", "N-Triples", None),
    ("somefile.nt.gz", "N-Triples", "gzip"),
    ("somefile.ntriples.gz", "N-Triples", "gzip"),
    ("somefile.nq", "N-Quads", None),
    ("somefile.nquads", "N-Quads", None),
    ("somefile.nq.gz", "N-Quads", "gzip"),
    ("somefile.nquads.gz", "N-Quads", "gzip"),
    ("somefile.nqx", "Extended N-Quads (with attributes)", None),
    ("somefile.nqx.gz", "Extended N-Quads (with attributes)", "gzip"),
    ("somefile.ttl", "Turtle", None),
    ("somefile.turtle", "Turtle", None),
    ("somefile.ttl.gz", "Turtle", "gzip"),
    ("somefile.turtle.gz", "Turtle", "gzip"),
    ("somefile.trig", "TriG", None),
    ("somefile.trig.gz", "TriG", "gzip"),
    ("somefile.trix", "TriX", None),
    ("somefile.trix.gz", "TriX", "gzip"),
    ("sOmEfIlE.tRiX.gZ", "TriX", "gzip"),
    ("somefile.kaboom", None, None),
])
def test_format_for_ext(filename, expected_format, expected_compression):
    actual_format, actual_compression = RDFFormat.format_for_file_name(filename)
    if actual_format is None:
        assert expected_format is None
    else:
        assert expected_format == actual_format.name
    assert expected_compression == actual_compression


@pytest.mark.parametrize("value, expected_text, expected_type", [
    (u'दुप', u'दुप', None), (b'test', u'test', None),
    (True, 'true', XMLSchema.BOOLEAN), (False, 'false', XMLSchema.BOOLEAN),
    (42, '42', XMLSchema.INTEGER), (42.000, '42.0', XMLSchema.DOUBLE),
    (Decimal('42.000'), '42.000', XMLSchema.DECIMAL),
    (datetime(1984, 8, 26, 10, 0, 5), '1984-08-26T10:00:05Z', XMLSchema.DATETIME),
    # TODO: Should we really be converting times to UTC?
    (datetime(1984, 8, 26, 12, 0, 5, tzinfo=MockTimezone('CEST', 2, 1)),
     '1984-08-26T10:00:05Z', XMLSchema.DATETIME),
    (datetime(1984, 8, 27, 1, 0, 5, tzinfo=MockTimezone('CEST', 2, 1)),
     '1984-08-26T23:00:05Z', XMLSchema.DATETIME),
    (date(1984, 8, 26), '1984-08-26', XMLSchema.DATE),
    (time(10, 0, 5), '10:00:05Z', XMLSchema.TIME),
    (time(12, 0, 5, tzinfo=MockTimezone('CEST', 2, 1)), '10:00:05Z', XMLSchema.TIME),
    (time(1, 0, 5, tzinfo=MockTimezone('CEST', 2, 1)), '23:00:05Z', XMLSchema.TIME),
    ([1, 2, 3], "[1, 2, 3]", None)])
def test_literals_from_python_values(value, expected_text, expected_type):
    literal = Literal(value)
    assert literal.label == expected_text
    # Well-known types are normalized, so it is safe to use the ``is`` operator here.
    assert literal.datatype is expected_type


@pytest.mark.skipif(sys.version_info >= (3,),
                    reason="Long type exists only on Python 2.")
def test_long_literal():
    # This behavior of using LONG as the datatype might be convenient in demos,
    # (you can do ``Literal(42L)``), but has a few drawbacks:
    #    - is incorrect: will map arbitrarily huge values to LONG, which is
    #      supposed to be 64 bit, while SMALLER values will become INTEGERS
    #      (which have no size limit).
    #    - In fact longs start at sys.maxint + 1 == 2 ** 63, i.e. as soon as
    #      it is no longer valid to hold them in xsd:longs.
    #    - Python 3 has no 'long' type.
    literal = Literal(84104105115032109097107101115032110111032115101110115101046)
    assert literal.datatype is XMLSchema.LONG


def test_true_literal():
    assert Literal(True).booleanValue()


def test_false_literal():
    assert not Literal(False).booleanValue()


def test_add_data_ascii(conn):
    base_uri = u'http://franz.com/'
    expected = conn.createURI(namespace=base_uri, localname='x')
    conn.addData(u'<x> <x> <x> .', base_uri=base_uri)
    actual = conn.getStatements(None, None, None, None).asList()
    assert len(actual) == 1
    assert actual[0].getSubject() == expected
    assert actual[0].getPredicate() == expected
    assert actual[0].getObject() == expected


def test_add_data_unicode(conn):
    """ See bug24405. """
    base_uri = u'http://franz.com/'
    expected = conn.createURI(namespace=base_uri, localname=u'दुप')
    conn.addData(u'<दुप> <दुप> <दुप> .', base_uri=base_uri)
    actual = conn.getStatements(None, None, None, None).asList()
    assert len(actual) == 1
    assert actual[0].getSubject() == expected
    assert actual[0].getPredicate() == expected
    assert actual[0].getObject() == expected


def test_get_statements_raw_value(conn):
    conn.addData('<ex://s1> <ex://p1> true .')
    conn.addData('<ex://s2> <ex://p2> false .')
    actual = conn.getStatements(None, None, False, None).asList()
    assert len(actual) == 1
    assert actual[0].getSubject() == conn.createURI("ex://s2")


# See spr43630
def test_unicode_literal_filter(conn, s, p):
    o = conn.createLiteral(u'<दुप>')
    conn.add(s, p, o)
    conn.getStatements(None, None, o)


def test_remove_statement(conn, s, p, o, x, g):
    conn.add(s, p, o, [g])
    conn.add(x, p, o, [g])
    conn.add(s, x, o, [g])
    conn.add(s, p, x, [g])
    conn.add(s, p, o, [x])
    stmt = Statement(s, p, o, g)
    conn.removeStatement(stmt, [g])
    remaining = conn.getStatements(None, None, None).asList()
    assert len(remaining) == 4
    for r in remaining:
        assert r.getSubject() != s or r.getPredicate() != p \
               or r.getObject() != o or r.getContext() != g


def test_no_suppression(conn, s, p, o, o2):
    assert conn.getDuplicateSuppressionPolicy() is None
    conn.add(s, p, o)
    conn.add(s, p, o)
    conn.add(s, p, o2)
    conn.commit()
    assert len(conn.getStatements(None, None, None)) == 3


def test_spo_suppression(conn, s, p, o, o2, g1, g2):
    conn.setDuplicateSuppressionPolicy("spo")
    assert conn.getDuplicateSuppressionPolicy() == "spo"
    conn.add(s, p, o, [g1])
    conn.add(s, p, o, [g1])
    conn.add(s, p, o2, [g1])
    conn.add(s, p, o, [g2])
    conn.commit()
    assert len(conn.getStatements(None, None, None)) == 2


def test_spog_suppression(conn, s, p, o, o2, g1, g2):
    conn.setDuplicateSuppressionPolicy("spog")
    assert conn.getDuplicateSuppressionPolicy() == "spog"
    conn.add(s, p, o, [g1])
    conn.add(s, p, o, [g1])
    conn.add(s, p, o2, [g1])
    conn.add(s, p, o, [g2])
    conn.commit()
    assert len(conn.getStatements(None, None, None)) == 3


def test_disable_suppression(conn, s, p, o, o2):
    conn.setDuplicateSuppressionPolicy("spo")
    conn.add(s, p, o)
    conn.add(s, p, o)
    conn.add(s, p, o2)
    conn.commit()
    conn.disableDuplicateSuppression()
    assert conn.getDuplicateSuppressionPolicy() is None
    conn.add(s, p, o)
    conn.add(s, p, o2)
    assert len(conn.getStatements(None, None, None)) == 4


def test_suppression_none(conn, s, p, o, o2):
    conn.setDuplicateSuppressionPolicy("spo")
    conn.add(s, p, o)
    conn.add(s, p, o)
    conn.add(s, p, o2)
    conn.commit()
    conn.setDuplicateSuppressionPolicy(None)
    assert conn.getDuplicateSuppressionPolicy() is None
    conn.add(s, p, o)
    conn.add(s, p, o2)
    assert len(conn.getStatements(None, None, None)) == 4


def test_switch_suppression(conn, s, p, o, o2, g1, g2):
    conn.setDuplicateSuppressionPolicy("spog")
    assert conn.getDuplicateSuppressionPolicy() == "spog"
    conn.add(s, p, o, [g1])
    conn.add(s, p, o, [g1])
    conn.add(s, p, o2, [g1])
    conn.add(s, p, o, [g2])
    conn.commit()
    conn.setDuplicateSuppressionPolicy("spo")
    assert conn.getDuplicateSuppressionPolicy() == "spo"
    conn.add(s, p, o)
    assert len(conn.getStatements(None, None, None)) == 3


def test_switch_suppression_before_commit(session, s, p, o, o2, g1, g2):
    session.setDuplicateSuppressionPolicy("spog")
    session.add(s, p, o, [g1])
    session.add(s, p, o, [g1])
    session.add(s, p, o2, [g1])
    session.add(s, p, o, [g2])
    session.setDuplicateSuppressionPolicy("spo")
    session.commit()
    assert len(session.getStatements(None, None, None)) == 2


def test_export_ntriples(conn):
    conn.addData("""<ex://s1> <ex://p1> <ex://o1> .
                    <ex://s2> <ex://p2> <ex://o2> .""")
    out = BytesIO()
    conn.getStatements(output=out, output_format=RDFFormat.NTRIPLES)
    result = sorted(out.getvalue().decode('utf-8').splitlines())
    assert result == [
        "<ex://s1> <ex://p1> <ex://o1> .",
        "<ex://s2> <ex://p2> <ex://o2> ."
    ]


def test_export_by_id(conn):
    conn.addData("""<ex://s1> <ex://p1> <ex://o1> .
                    <ex://s2> <ex://p2> <ex://o2> .""")
    ids = [stmt.getTripleID()
           for stmt in conn.getStatements(subject=conn.createURI('ex://s2'), tripleIDs=True)]

    out = BytesIO()
    conn.getStatementsById(ids, output=out, output_format=RDFFormat.NTRIPLES)
    result = sorted(out.getvalue().decode('utf-8').splitlines())
    assert result == [
        "<ex://s2> <ex://p2> <ex://o2> ."
    ]


def test_export_query_result(conn):
    conn.addData("""<ex://s1> <ex://p1> <ex://o1> .
                    <ex://s2> <ex://p2> <ex://o2> .""")
    query_string = "select ?s { ?s <ex://p1> ?o }"
    query = conn.prepareTupleQuery(QueryLanguage.SPARQL, query_string)
    out = BytesIO()
    query.evaluate(output=out, output_format=TupleFormat.CSV)
    result = re.sub(r'\s', '', out.getvalue().decode('utf-8'))
    assert result == 's"ex://s1"'


def test_export_construct(conn):
    query = conn.prepareGraphQuery(QueryLanguage.SPARQL, """
        CONSTRUCT {
          <ex://s1> <ex://p1> <ex://o1> .
          <ex://s2> <ex://p2> <ex://o2> .
        } WHERE {}""")
    out = BytesIO()
    query.evaluate(output=out, output_format=RDFFormat.NTRIPLES)
    result = sorted(out.getvalue().decode('utf-8').splitlines())
    assert result == [
        "<ex://s1> <ex://p1> <ex://o1> .",
        "<ex://s2> <ex://p2> <ex://o2> ."
    ]

@min_version(6, 6)
def test_export_query_result_tsv(conn):
    base_uri = u'http://franz.com/'
    s = conn.createURI(namespace=base_uri, localname='s')
    p = conn.createURI(namespace=base_uri, localname='p')
    o = conn.createURI(namespace=base_uri, localname='aa \t \n \r \\ bb')
    conn.add(s, p, o)
    conn.commit()
    query_string = "select ?o { ?s ?p ?o }"
    query = conn.prepareTupleQuery(QueryLanguage.SPARQL, query_string)
    out = BytesIO()
    query.evaluate(output=out, output_format=TupleFormat.TSV)
    assert out.getvalue().decode('utf-8') == '?o\n<http://franz.com/aa \\t \\n \\r \\\\ bb>\n'


def test_datetime_value_tz():
    lit = Literal('1984-08-26T10:00:05+02:00', XMLSchema.DATETIME)
    dt = lit.datetimeValue()
    assert dt.utcoffset() == timedelta(0, 7200)


def test_time_value_tz():
    lit = Literal('1984-08-26T10:00:05+02:00', XMLSchema.DATETIME)
    t = lit.datetimeValue()
    assert t.utcoffset() == timedelta(0, 7200)


def test_datetime_tz_roundtrip(conn, s, p):
    dt = conn.createLiteral('1984-08-26T10:00:05+02:00', XMLSchema.DATETIME)
    conn.addTriple(s, p, dt)
    actual = conn.getStatements().asList()[0].getObject()
    assert actual == dt


def test_time_tz_roundtrip(conn, s, p):
    t = conn.createLiteral('10:00:05+02:00', XMLSchema.TIME)
    conn.addTriple(s, p, t)
    actual = conn.getStatements().asList()[0].getObject()
    assert actual == t


def test_invalid_query(conn):
    query = conn.prepareTupleQuery(QueryLanguage.SPARQL,
                                   '"; DROP TABLE students;')
    with open(os.devnull, 'w') as out:
        with pytest.raises(RequestError):
            query.evaluate(output=out)


def test_default_rdf_format(conn):
    out = BytesIO()
    query = conn.prepareGraphQuery(QueryLanguage.SPARQL,"""
        CONSTRUCT { <ex://s> <ex://p> <ex://o> } WHERE {}""")
    query.evaluate(output=out)
    assert out.getvalue()


def test_default_tuple_format(conn):
    out = BytesIO()
    query = conn.prepareTupleQuery(QueryLanguage.SPARQL,"""
        SELECT ?x { BIND(42 as ?x) }""")
    query.evaluate(output=out)
    assert out.getvalue()


def test_output_to_true(capfd):
    with output_to(True) as f:
        f.write(b'test')
    out, err = capfd.readouterr()
    assert out == 'test'


def test_output_to_stderr(capfd):
    with output_to(2) as f:
        f.write(b'test')
    out, err = capfd.readouterr()
    assert err == 'test'


def test_export_to_true(capfd, conn, s, p, o):
    conn.add(s, p, o)
    conn.getStatements(output=True)
    out, err = capfd.readouterr()
    assert out.strip() == "<ex://s> <ex://p> <ex://o> ."


def test_triple_len(s, p, o):
    assert len(Statement(s, p, o)) == 3


def test_quad_len(s, p, o, g):
    assert len(Statement(s, p, o, g)) == 4


def test_stmt_str(ex):
    # This used to raise an error, see bug25079
    assert 'ex' in str(Statement(ex.s, ex.p, ex.o))


def test_statement_from_strings_subject(ex):
    stmt = Statement('<ex://s>', '<ex://p>', '<ex://o>')
    assert stmt.getSubject() == ex.s


def test_statement_from_strings_predicate(ex):
    stmt = Statement('<ex://s>', '<ex://p>', '<ex://o>')
    assert stmt.getPredicate() == ex.p


def test_statement_from_strings_object(ex):
    stmt = Statement('<ex://s>', '<ex://p>', '<ex://o>')
    assert stmt.getObject() == ex.o


def test_triple_ne_quad(ex):
    s1 = Statement(ex.s, ex.p, ex.o)
    s2 = Statement(ex.s, ex.p, ex.o, ex.g)
    assert s1 != s2


def test_add_triples_statements(conn, ex):
    s = Statement(ex.s, ex.p, ex.o)
    conn.addTriples([s])
    assert conn.getStatements().asList() == [s]


def test_conn_namespace(conn):
    ex = conn.namespace('http://franz.com/example/')
    assert ex.foo == conn.createURI('http://franz.com/example/foo')

def test_conn_namespace_array(conn):
    ex = conn.namespace('http://franz.com/example/')
    assert ex['bar'] == conn.createURI('http://franz.com/example/bar')

def test_conn_namespace_call(conn):
    ex = conn.namespace('http://franz.com/example/')
    assert ex('baz') == conn.createURI('http://franz.com/example/baz')

def test_add_with_single_context(conn, s, p, o, g):
    conn.addTriples([(s, p, o)], context=g)
    assert get_statements(conn) == [[s, p, o, g]]


def test_add_with_two_contexts(conn, ex):
    conn.addTriples([(ex.s, ex.p, ex.o)], context=[ex.g1, ex.g2])
    assert get_statements(conn) == [[ex.s, ex.p, ex.o, ex.g1],
                                    [ex.s, ex.p, ex.o, ex.g2]]


def test_add_quad_overrides_contexts(conn, ex):
    conn.addTriples([(ex.s, ex.p, ex.o, ex.g)], context=[ex.g1, ex.g2])
    assert get_statements(conn) == [[ex.s, ex.p, ex.o, ex.g]]


# A helper function that makes a request and intercepts
# the repl header. It also sorts all the settings
# in the header to make comparisons easier.
def get_repl_header(conn, ex, mocker):
    req = mocker.spy(franz.miniclient.request, 'makeRequest')
    conn.addTriple(ex.s, ex.p, ex.o)
    # not implemented in Python 2...
    # req.assert_called()
    for _args, kwargs in req.call_args_list:
        headers = normalize_headers(kwargs.get('headers', {})) or {}
        return normalize_repl_header(headers.get('x-repl-settings'))


# Sort values in a repl header
def normalize_repl_header(header):
    if header is None:
        return None
    return ' '.join(sorted(header.split()))


def test_default_get_transaction_settings(conn):
    assert conn.getTransactionSettings() is None


def test_set_durability(conn):
    conn.setTransactionSettings(durability=42)
    assert conn.getTransactionSettings().durability == 42


def test_set_durability_does_not_change_latency(conn):
    conn.setTransactionSettings(transaction_latency_count=7)
    conn.setTransactionSettings(durability=42)
    assert conn.getTransactionSettings().transaction_latency_count == 7


def test_set_settings_object(conn):
    conn.setTransactionSettings(durability=42)
    conn.setTransactionSettings(TransactionSettings(transaction_latency_count=7))
    assert conn.getTransactionSettings().transaction_latency_count == 7
    assert conn.getTransactionSettings().durability is None


def test_settings_object_and_kwargs(conn):
    conn.setTransactionSettings(durability=42)
    conn.setTransactionSettings(
        TransactionSettings(transaction_latency_count=7),
        transaction_latency_timeout=2000)
    assert conn.getTransactionSettings().transaction_latency_count == 7
    assert conn.getTransactionSettings().transaction_latency_timeout == 2000
    assert conn.getTransactionSettings().durability is None


def test_settings_object_and_kwargs_overwrite(conn):
    conn.setTransactionSettings(
        TransactionSettings(durability=7), durability=42)
    assert conn.getTransactionSettings().durability == 42


def test_default_transaction_settings(conn, ex, mocker):
    assert get_repl_header(conn, ex, mocker) is None


def test_temporary_settings_object(conn):
    new = TransactionSettings(durability=7)
    with conn.temporaryTransactionSettings(new):
        assert conn.getTransactionSettings() == new


def test_temporary_settings_object_restore(conn):
    old = TransactionSettings(durability=42)
    new = TransactionSettings(durability=7)
    conn.setTransactionSettings(old)
    with conn.temporaryTransactionSettings(new):
        pass
    assert conn.getTransactionSettings() == old


def test_temporary_settings_kwargs(conn):
    with conn.temporaryTransactionSettings(durability=7):
        assert conn.getTransactionSettings().durability == 7


def test_temporary_settings_kwargs_restore(conn):
    old = TransactionSettings(durability=42)
    conn.setTransactionSettings(old)
    with conn.temporaryTransactionSettings(durability=7):
        pass
    assert conn.getTransactionSettings() == old


def test_temporary_settings_kwargs_overwrite(conn):
    old = TransactionSettings(durability=42, transaction_latency_count=10)
    conn.setTransactionSettings(old)
    with conn.temporaryTransactionSettings(durability=7):
        assert conn.getTransactionSettings().durability == 7
        assert conn.getTransactionSettings().transaction_latency_count == 10


def test_temporary_settings_object_and_kwargs_overwrites_all(conn):
    old = TransactionSettings(durability=42)
    new = TransactionSettings()
    conn.setTransactionSettings(old)
    with conn.temporaryTransactionSettings(new, transaction_latency_count=7):
        assert conn.getTransactionSettings().durability is None


def test_temporary_settings_object_and_kwargs(conn):
    old = TransactionSettings(durability=42)
    new = TransactionSettings(durability=7)
    conn.setTransactionSettings(old)
    with conn.temporaryTransactionSettings(new, transaction_latency_count=100):
        assert conn.getTransactionSettings().durability == 7
        assert conn.getTransactionSettings().transaction_latency_count == 100
    assert conn.getTransactionSettings() == old


def test_temporary_settings_nonlocal_return(conn):
    old = TransactionSettings(durability=42)
    new = TransactionSettings(durability=7)
    conn.setTransactionSettings(old)
    while True:
        with conn.temporaryTransactionSettings(new):
            break
    assert conn.getTransactionSettings() == old


def test_set_durability_header(conn, ex, mocker):
    conn.setTransactionSettings(durability=7)
    assert get_repl_header(conn, ex, mocker) == 'durability=7'


def test_set_durability_and_latency_count(conn, ex, mocker):
    conn.setTransactionSettings(durability=7, transaction_latency_count=3)
    settings = get_repl_header(conn, ex, mocker)
    assert settings == 'durability=7 transactionLatencyCount=3'


def test_set_all_transaction_settings(conn, ex, mocker):
    conn.setTransactionSettings(distributed_transaction_timeout=1,
                                durability=2, transaction_latency_count=3,
                                transaction_latency_timeout=4)
    settings = get_repl_header(conn, ex, mocker)
    assert settings == ' '.join(('distributedTransactionTimeout=1',
                                 'durability=2',
                                 'transactionLatencyCount=3',
                                 'transactionLatencyTimeout=4'))


def test_latency_timeout_timedelta(conn, ex, mocker):
    conn.setTransactionSettings(transaction_latency_timeout=timedelta(hours=1))
    assert get_repl_header(conn, ex, mocker) == 'transactionLatencyTimeout=3600'


def test_timeout_timedelta(conn, ex, mocker):
    conn.setTransactionSettings(distributed_transaction_timeout=timedelta(hours=1))
    assert get_repl_header(conn, ex, mocker) == 'distributedTransactionTimeout=3600'


def test_commit_settings(conn, ex, mocker):
    req = mocker.spy(franz.miniclient.request, 'makeRequest')
    conn.addTriple(ex.s, ex.p, ex.o)
    conn.commit(durability='quorum')
    # Not available in Python 2
    # req.assert_called()
    args, kwargs = req.call_args
    headers = normalize_headers(kwargs.get('headers', {})) or {}
    header = normalize_repl_header(headers.get('x-repl-settings'))
    assert header == 'durability=quorum'


def test_define_attribute(conn):
    attr = AttributeDefinition(
        name='test', allowed_values=['a', 'b', 'c'],
        minimum_number=0, maximum_number=1)
    conn.setAttributeDefinition(attr)
    actual = conn.getAttributeDefinition('test')
    assert actual.name == 'test'
    assert sorted(actual.allowed_values) == ['a', 'b', 'c']
    assert actual.ordered is False
    assert actual.minimum_number == 0
    assert actual.maximum_number == 1


def test_get_set_attribute_filter_str(conn, attr):
    conn.setAttributeFilter('(empty triple.test)')
    assert conn.getAttributeFilter() == '(empty triple.test)'


def test_attribute_filter_works(conn, ex, attr):
    conn.addTriple(ex.s1, ex.p1, ex.o1, attributes={'test': 'a'})
    conn.addTriple(ex.s2, ex.p2, ex.o2, attributes={'test': 'b'})
    conn.addTriple(ex.s3, ex.p3, ex.o3, attributes={'test': 'c'})
    conn.setAttributeFilter(TripleAttribute.test == 'b')
    assert get_statements(conn) == [[ex.s2, ex.p2, ex.o2, None]]
    

def test_attribute_filter_works_with_output(conn, ex, attr):
    conn.addTriple(ex.s1, ex.p1, ex.o1, attributes={'test': 'a'})
    conn.addTriple(ex.s2, ex.p2, ex.o2, attributes={'test': 'b'})
    conn.addTriple(ex.s3, ex.p3, ex.o3, attributes={'test': 'c'})
    conn.setAttributeFilter(TripleAttribute.test == 'b')
    buf = io.BytesIO()
    conn.executeTupleQuery('select ?s { ?s ?p ?o }', 
                           output=buf, output_format=TupleFormat.CSV)
    assert buf.getvalue().strip().split() == [b's', b'"ex://s2"']


def test_attribute_filter_with_user_attributes(conn, ex, attr):
    conn.addTriple(ex.s1, ex.p1, ex.o1, attributes={'test': 'a'})
    conn.addTriple(ex.s2, ex.p2, ex.o2, attributes={'test': 'b'})
    conn.addTriple(ex.s3, ex.p3, ex.o3, attributes={'test': 'c'})
    conn.setUserAttributes({'test': 'b'})
    conn.setAttributeFilter(TripleAttribute.test == UserAttribute.test)
    assert get_statements(conn) == [[ex.s2, ex.p2, ex.o2, None]]


@pytest.mark.parametrize("attr_set, expected_text", [
    ('raw-value', '"raw-value"'),
    ('value-with-"-inside', '"value-with-\\"-inside"'),
    (['a', 'list', 'of', 'strings'], '("a" "list" "of" "strings")'),
    (UserAttribute.test, 'user.test'),
    (TripleAttribute.level, 'triple.level')])
def test_attr_set_expr(attr_set, expected_text):
    assert expected_text == attribute_set_to_expr(attr_set)


@pytest.mark.parametrize("attr_filter, expected_text", [
    ('(empty ("raw" "string"))', '(empty ("raw" "string"))'),
    (And('(and)', '(or)'), '(and (and) (or))'),
    (Or('(and)', '(or)'), '(or (and) (or))'),
    (And('(and)',Or()), '(and (and) (or))'),
    (Or(And(), '(or)'), '(or (and) (or))'),
    (Not('(and)'), '(not (and))'),
    (Not(Not(Not(And()))), '(not (not (not (and))))'),
    (Empty([]), '(empty ())'),
    (Empty(TripleAttribute.restricted), '(empty triple.restricted)'),
    (Overlap([], ["x"]), '(overlap () ("x"))'),
    (Overlap(UserAttribute.flags, TripleAttribute.flags),
     '(overlap user.flags triple.flags)'),
    (Subset([], ["x"]), '(subset () ("x"))'),
    (Subset(UserAttribute.flags, TripleAttribute.flags),
     '(subset user.flags triple.flags)'),
    (Superset([], ["x"]), '(superset () ("x"))'),
    (Superset(UserAttribute.flags, TripleAttribute.flags),
     '(superset user.flags triple.flags)'),
    (Equal(["y"], ["x"]), '(equal ("y") ("x"))'),
    (Equal(UserAttribute.color, TripleAttribute.color),
     '(equal user.color triple.color)'),
    (Lt(UserAttribute.level, TripleAttribute.level),
     '(attribute-set< user.level triple.level)'),
    (Le(UserAttribute.level, TripleAttribute.level),
     '(attribute-set<= user.level triple.level)'),
    (Eq(UserAttribute.level, TripleAttribute.level),
     '(attribute-set= user.level triple.level)'),
    (Ge(UserAttribute.level, TripleAttribute.level),
     '(attribute-set>= user.level triple.level)'),
    (Gt(UserAttribute.level, TripleAttribute.level),
     '(attribute-set> user.level triple.level)'),
    (UserAttribute.x < TripleAttribute.x, '(attribute-set< user.x triple.x)'),
    (UserAttribute.x <= TripleAttribute.x, '(attribute-set<= user.x triple.x)'),
    (UserAttribute.x == TripleAttribute.x, '(equal user.x triple.x)'),
    (UserAttribute.x > TripleAttribute.x, '(attribute-set> user.x triple.x)'),
    (UserAttribute.x >= TripleAttribute.x, '(attribute-set>= user.x triple.x)'),
    (UserAttribute.x << TripleAttribute.x, '(subset user.x triple.x)'),
    (UserAttribute.x >> TripleAttribute.x, '(superset user.x triple.x)'),
    ('a' < TripleAttribute.x, '(attribute-set> triple.x "a")'),
    ('a' <= TripleAttribute.x, '(attribute-set>= triple.x "a")'),
    ('a' == TripleAttribute.x, '(equal triple.x "a")'),
    ('a' > TripleAttribute.x, '(attribute-set< triple.x "a")'),
    ('a' >= TripleAttribute.x, '(attribute-set<= triple.x "a")'),
    ('a' << TripleAttribute.x, '(subset "a" triple.x)'),
    ('a' >> TripleAttribute.x, '(superset "a" triple.x)'),
    (Empty([]) & '(or)', '(and (empty ()) (or))'),
    (Empty([]) & Or(), '(and (empty ()) (or))'),
    (Empty([]) | '(and)', '(or (empty ()) (and))'),
    (Empty([]) | And(), '(or (empty ()) (and))'),
    (~Empty([]), '(not (empty ()))')
])
def test_attr_filter_to_expr(attr_filter, expected_text):
    assert expected_text == attribute_filter_to_expr(attr_filter)


def normalize_attributes(d):
    """
    Sort returned attribute values.
    """
    if d is None:
        return None
    r = {}
    for key, value in d.items():
        if isinstance(value, list):
            r[key] = sorted(value)
        else:
            r[key] = value
    return r


def get_triple_attributes(conn, s, p, o, g=None):
    if g is None:
        qt = '''select ?a { 
                   ?a <http://franz.com/ns/allegrograph/6.2.0/attributes> (?s ?p ?o) . 
        }'''
    else:
        qt = '''select ?a { 
                   graph ?g {
                       ?a <http://franz.com/ns/allegrograph/6.2.0/attributes> (?s ?p ?o) .
                   } 
        }'''
    query = conn.prepareTupleQuery(query=qt)
    query.setBinding('s', s)
    query.setBinding('p', p)
    query.setBinding('o', o)
    if g is not None:
        query.setBinding('g', g)
    with query.evaluate() as result:
        return [normalize_attributes(json.loads(b['a'].label or 'null'))
                for b in result
                if b['a'] is not None]


def test_add_data_with_attributes(conn, ex, attr):
    conn.addData('<ex://s> <ex://p> <ex://o> .', attributes={'test': 'c'})
    assert [{'test': 'c'}] == get_triple_attributes(conn, ex.s, ex.p, ex.o)


def test_add_tuples_with_attributes(conn, ex, attr):
    conn.addTriples([[ex.s1, ex.p1, ex.o1], [ex.s2, ex.p2, ex.o2]],
                    attributes={'test': 'c'})
    assert [{'test': 'c'}] == get_triple_attributes(conn, ex.s1, ex.p1, ex.o1)
    assert [{'test': 'c'}] == get_triple_attributes(conn, ex.s2, ex.p2, ex.o2)


def test_add_quints(conn, ex, attr):
    conn.addTriples([[ex.s1, ex.p1, ex.o1, None, {'test': 'a'}],
                    [ex.s2, ex.p2, ex.o2, None, {'test': 'b'}]])
    assert [{'test': 'a'}] == get_triple_attributes(conn, ex.s1, ex.p1, ex.o1)
    assert [{'test': 'b'}] == get_triple_attributes(conn, ex.s2, ex.p2, ex.o2)


def test_add_quints_multiple_values(conn, ex, attr2):
    conn.addTriples([[ex.s1, ex.p1, ex.o1, None, {'test2': ['a', 'b']}],
                     [ex.s2, ex.p2, ex.o2, None, {'test2': ['b', 'c']}]])
    assert [{'test2': ['a', 'b']}] == get_triple_attributes(conn, ex.s1, ex.p1, ex.o1)
    assert [{'test2': ['b', 'c']}] == get_triple_attributes(conn, ex.s2, ex.p2, ex.o2)


def test_add_quints_with_attributes(conn, ex, attr):
    conn.addTriples([[ex.s1, ex.p1, ex.o1, None, {'test': 'a',}],
                     [ex.s2, ex.p2, ex.o2],
                     [ex.s3, ex.p3, ex.o3, None, None],
                     [ex.s4, ex.p4, ex.o4, None, {}]],
                    attributes={'test': 'c'})
    assert [{'test': 'a'}] == get_triple_attributes(conn, ex.s1, ex.p1, ex.o1)
    assert [{'test': 'c'}] == get_triple_attributes(conn, ex.s2, ex.p2, ex.o2)
    assert [{'test': 'c'}] == get_triple_attributes(conn, ex.s3, ex.p3, ex.o3)
    assert [] == get_triple_attributes(conn, ex.s4, ex.p4, ex.o4)


def test_set_user_data(server):
    server.setUserData('testKey', 'Hello, "World"!')
    try:
        assert server.getUserData('testKey') == 'Hello, "World"!'
    finally:
        server.deleteUserData('testKey')


def test_user_data_not_found(server):
    assert server.getUserData('nope') is None


def test_del_user_data(server):
    server.setUserData('testKey', 'Hello, "World"!')
    server.deleteUserData('testKey')
    assert server.getUserData('testKey') is None


def test_user_data_fixture(server, user_data):
    user_data['testKey2'] = 'hello!'
    assert server.getUserData('testKey2') == 'hello!'
    del user_data['testKey2']
    assert server.getUserData('testKey2') is None


def test_http_sanity(http_server):
    http_server.publish('/', b'Hello!')
    r = requests.get(http_server.url('/'))
    assert r.text == 'Hello!'


def test_remote_http_sanity(remote_http_server):
    remote_http_server.publish('/', 'Hello!')
    assert remote_http_server.send_request('/') == 'Hello!'


@min_version(6, 5)
def test_add_json_ld(conn, example):
    conn.addData('''{
      "@context": {
        "@vocab": "http://franz.com/example/"
      },
      "@id": "subject",
      "predicate": { "@id": "object" }
    }''', rdf_format=RDFFormat.JSONLD, base_uri=example('').uri)
    assert get_statements(conn) == [[example.subject, example.predicate, example.object, None]]


@min_version(6, 5)
def test_add_json_from_dict(conn, example):
    conn.addData({
      "@context": {
        "@vocab": "http://franz.com/example/"
      },
      "@id": "subject",
      "predicate": {"@id": "object"}
    }, base_uri=example('').uri)
    assert get_statements(conn) == [[example.subject, example.predicate, example.object, None]]


@min_version(6, 5)
def test_add_json_ld_with_context(conn, example):
    conn.addData('''{
      "@id": "subject",
      "predicate": { "@id": "object" }
    }''', rdf_format=RDFFormat.JSONLD, json_ld_context={
        "@vocab": "http://franz.com/example/"
    }, base_uri=example('').uri)
    assert get_statements(conn) == [[example.subject, example.predicate, example.object, None]]


@min_version(6, 5)
def test_add_json_ld_with_external_context(conn, example, remote_http_server):
    url = remote_http_server.publish('/ctx.json', '{"@context": {"@vocab":"http://franz.com/example/"}}')
    conn.addData('''{
      "@id": "subject",
      "predicate": { "@id": "http://franz.com/example/object" }}''',
                 rdf_format=RDFFormat.JSONLD,
                 json_ld_context=url,
                 allow_external_references=True,
                 base_uri=example('').uri)
    assert get_statements(conn) == [[example.subject, example.predicate, example.object, None]]


@min_version(6, 5)
def test_add_json_ld_with_external_context_inside(conn, example, remote_http_server):
    url = remote_http_server.publish('/ctx.json', json.dumps({
       "@context": {
           "predicate": {
               "@id": "http://franz.com/example/predicate",
               "@type": "@id"
           }
       }
    }))
    conn.addData('''{
      "@context": %s,
      "@id": "http://franz.com/example/subject",
      "predicate": { "@id": "http://franz.com/example/object" }}''' % json.dumps(url),
                 rdf_format=RDFFormat.JSONLD,
                 allow_external_references=True)
    assert get_statements(conn) == [[example.subject, example.predicate, example.object, None]]


@min_version(6, 5)
def test_add_json_ld_keep_source(conn):
    src = '''{"ex://p": "o"}'''
    conn.addData(src, rdf_format=RDFFormat.JSONLD, json_ld_store_source=True)
    statements = get_statements(conn)
    assert 2 == len(statements)
    assert statements[0][2].label == src or statements[1][2].label == src

@min_version(6, 6)
def test_add_json_ld_with_root_graph(conn):
    src = '''{"ex://p": "o"}'''
    conn.addData(src, rdf_format=RDFFormat.JSONLD, context=':root')
    statements = get_statements(conn)
    assert 1 == len(statements)
    assert statements[0][0] == statements[0][3]
    

@min_version(6, 5)
def test_add_json_ld_empty_dict(conn):
    conn.addData({})
    assert [] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_simple_dict(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        'p': {'@id': 'o'}
    })
    assert [[example.s, example.p, example.o, None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_uri_key(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        example.p: {'@id': 'o'}
    })
    assert [[example.s, example.p, example.o, None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_uri_values(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': example.s,
        'p': {'@id': example.o}
    })
    assert [[example.s, example.p, example.o, None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_literal_value(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        example.p: Literal('o')
    })
    assert [[example.s, example.p, Literal('o'), None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_typed_literal_value(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        example.p: Literal('oooo', XMLSchema.BASE64BINARY)
    })
    assert [[example.s,
             example.p,
             Literal('oooo', XMLSchema.BASE64BINARY),
             None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_lang_literal_value(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        # A bathtub in Sindarin
        example.p: Literal('🛀', language='sjn')
    })
    assert [[example.s, example.p, Literal('🛀', language='sjn'), None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_integer_literal_value(conn, example):
    conn.addData({
        '@context': {
            '@vocab': 'http://franz.com/example/',
            '@base': 'http://franz.com/example/'
        },
        '@id': 's',
        example.p: Literal(42)
    })
    assert [[example.s, example.p, Literal(42), None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_dict_with_terms(conn, ex):
    conn.addData({
        '@id': ex.s,
        ex.p: ex.o
    })
    assert [[ex.s, ex.p, ex.o, None]] == get_statements(conn)


@min_version(6, 5)
def test_add_json_ld_list(conn, ex):
    conn.addData([{
        '@id': ex.s,
        ex.p: ex.o1
    }, {
        '@id': ex.s,
        ex.p: ex.o2
    }])
    assert [[ex.s, ex.p, ex.o1, None], 
            [ex.s, ex.p, ex.o2, None]] == get_statements(conn)


def test_uri_canonical_true():
    uri1 = URI('ex://test')
    uri2 = URI(namespace='ex://', localname='test')
    assert uri1 is uri2


def test_uri_canonical_false():
    uri1 = URI('ex://test', canonical=False)
    uri2 = URI('ex://test', canonical=False)
    assert uri1 is not uri2


def test_uri_canonicalize():
    uri1 = URI('ex://test', canonical=False)
    uri2 = URI(uri1)
    uri3 = URI('ex://test')
    assert uri2 is not uri1
    assert uri2 is uri3


def test_prolog_default_graph(conn, ex):
    conn.addTriple(ex.s, ex.p, ex.o)
    query = '(select (?s ?p ?o ?g) (q ?s ?p ?o ?g))'
    with conn.executeTupleQuery(query, QueryLanguage.PROLOG) as result:
        bindings = normalize_query_result(result)
    assert bindings == [[ex.s, ex.p, ex.o, None]]

def test_warmup(conn):
    """
    test warmup with various arguments.  we only verify that the HTTP
    call is valid and not that the warmup occured.
    """

    conn.warmup()
    conn.warmup(includeStrings=True)
    conn.warmup(includeStrings=False, includeTriples=False)
    conn.warmup(indices='spogi')
    conn.warmup(indices=['spogi', 'posgi'], includeTriples=True)

def test_sparql_query_metadata_select(conn):
    query = conn.prepareTupleQuery(QueryLanguage.SPARQL,"""
        SELECT (count(*) as ?count) { ?s ?p ?o }""")
    result = query.evaluate()
    md = result.getMetadata()
    assert md
    assert md['time']['total'] > 0

@min_version(7, 1)
def test_query_option_management(conn):
    """
    Test query option management.
    """
    # Default set of query options is empty.
    conn.clearQueryOptions()
    assert len(conn.getQueryOptions()) == 0

    assert_raises(RequestError, lambda n: conn.setQueryOption(n, ''),
                  'unknownQueryOption')

    test_options = {
        'logLineLength': 100,
        'authorizationBasic': '{}:{}'.format(USER, PASSWORD)
    }

    for name, value in iteritems(test_options):
        conn.setQueryOption(name, value)
    assert len(conn.getQueryOptions()) == len(test_options)

    for name, value in iteritems(test_options):
        assert value == conn.getQueryOption(name)

    # Try setting a query option that is already set.
    for name, value in iteritems(test_options):
        conn.setQueryOption(name, value)
    assert len(conn.getQueryOptions()) == len(test_options)

    # Remove one of the options.
    conn.removeQueryOption('logLineLength')
    assert len(conn.getQueryOptions()) == len(test_options) - 1
    assert_raises(RequestError, conn.getQueryOption, 'logLineLength')

    # Test clearing all query options.
    conn.clearQueryOptions()
    assert len(conn.getQueryOptions()) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from future.builtins import object
from past.builtins import unicode

from future.utils import iteritems

from ..model.value import URI 

NS = "http://www.w3.org/2000/01/rdf-schema#"

class RDFS(object):
    """
    A 'static' class containing useful RDFS URIs.
    """

    NAMESPACE = NS
    RESOURCE = URI(namespace=NS, localname="Resource")
    LITERAL = URI(namespace=NS, localname="Literal")
    CLASS = URI(namespace=NS, localname="Class")
    SUBCLASSOF = URI(namespace=NS, localname="subClassOf")
    SUBPROPERTYOF = URI(namespace=NS, localname="subPropertyOf")
    DOMAIN = URI(namespace=NS, localname="domain")
    RANGE = URI(namespace=NS, localname="range")
    COMMENT = URI(namespace=NS, localname="comment")
    LABEL = URI(namespace=NS, localname="label")
    DATATYPE = URI(namespace=NS, localname="Datatype")
    CONTAINER = URI(namespace=NS, localname="Container")
    MEMBER = URI(namespace=NS, localname="member")
    ISDEFINEDBY = URI(namespace=NS, localname="isDefinedBy")
    SEEALSO = URI(namespace=NS, localname="seeAlso")
    CONTAINERMEMBERSHIPPROPERTY = URI(namespace=NS,
	localname="ContainerMembershipProperty")





